import struct
import hashlib
import os
import numpy as np

HASH_LENGTH = 24


def file_md5(filename):
    hash_md5 = hashlib.md5()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()[:HASH_LENGTH]


def join_text_chunks(chunks):
    return "".join(chunks)


# Filenames for generated files
def get_converted_pdf_txt_filename(md5):
    return f"{md5}.pdf.txt"


def get_pdf_positions_filename(md5):
    return f"{md5}.pdf.positions.json"


def get_tokens_filename(md5, config_hash):
    return f"{md5}.{config_hash}.tokens.json"


def get_embeddings_filename(md5, config_hash, size, offset, rewind):
    return f"{md5}.{config_hash}.{size}_{offset}_{rewind}.embeddings"


def get_annoy_filename(md5, config_hash, size, offset, rewind, num_trees):
    return f"{md5}.{config_hash}.{size}_{offset}_{rewind}.{num_trees}t.annoy"


def get_config_filename(md5, config_hash):
    return f"{md5}.{config_hash}.config.json"


def write_embedding(file, embedding, num_dimensions):
    # Write float-encoded embeddings
    for i in range(num_dimensions):
        file.write(struct.pack("f", embedding[i]))
    file.flush()


def read_embedding(chunk, num_dimensions):
    # Read float-encoded embeddings
    embedding = []
    for i in range(num_dimensions):
        embedding.append(struct.unpack("f", chunk[i * 4 : (i + 1) * 4])[0])
    return embedding


def write_annoy_db(filename, num_dimensions, embeddings, num_trees):
    # Import annoy here so that it's not required for the CLI
    from annoy import AnnoyIndex

    dbs = []
    db = AnnoyIndex(num_dimensions, "angular")
    for i, embedding in enumerate(embeddings):
        db.add_item(i, embedding)
    db.build(num_trees)
    db.save(filename)
    dbs.append(db)

    return dbs


def load_annoy_db(filename, num_dimensions):
    # Import annoy here so that it's not required for the CLI
    from annoy import AnnoyIndex

    db = AnnoyIndex(num_dimensions, "angular")
    db.load(filename)
    return db


def get_num_annoy_embeddings(annoy_filename, num_dimensions):
    return load_annoy_db(annoy_filename, num_dimensions).get_n_items()


def safe_remove(filename):
    try:
        os.remove(filename)
    except FileNotFoundError:
        pass


def get_num_embeddings(embeddings_filename, num_dimensions):
    # Get the file size
    with open(embeddings_filename, "rb") as f:
        f.seek(0, 2)
        file_size = f.tell()

    # Calculate the number of embeddings
    return file_size // (num_dimensions * 4)


def read_embeddings_file(embeddings_filename, num_dimensions, capacity):
    # Calculate the number of embeddings
    num_embeddings = min(
        get_num_embeddings(embeddings_filename, num_dimensions), capacity
    )

    # Change the file size to the expected size
    with open(embeddings_filename, "ab") as f:
        f.truncate(num_embeddings * num_dimensions * 4)

    if num_embeddings == 0:
        return np.zeros((capacity, num_dimensions), dtype="float32"), 0

    # Memory map the file
    read_embeddings = np.memmap(
        embeddings_filename,
        dtype="float32",
        mode="r",
        shape=(num_embeddings, num_dimensions),
    )

    # Create an array with shape (capacity, num_dimensions) filled with 0s
    embeddings = np.zeros((capacity, num_dimensions), dtype="float32")

    # Copy the original embeddings into the new array
    embeddings[:num_embeddings] = read_embeddings[:num_embeddings]

    return embeddings, num_embeddings


def get_offsets(doc_size, windows):
    num_tokens = 0

    offsets = []

    for size, offset, rewind in windows:
        sub_offsets = []
        x = 0
        if offset > 0:
            sub_offsets.append([0, offset])
            num_tokens += offset
            x = offset
        else:
            x = rewind

        while x < doc_size:
            x -= rewind
            sub_offsets.append([x, min(x + size, doc_size)])
            num_tokens += min(x + size, doc_size) - x
            x += size

        offsets.append(sub_offsets)

    return offsets, num_tokens


def sort_results(results, reverse):
    # Get average distance per result
    avg_distances = []
    for result in results:
        avg_distances.append(np.mean([item["distance"] for item in result[1]]))

    # Sort results by average distance
    return {
        "results": [x for _, x in sorted(zip(avg_distances, results), reverse=reverse)],
        "sort": "desc" if reverse else "asc",
    }



import pypdfium2 as pdfium
from threading import Lock
import json
import os
from tqdm import tqdm
from .util import get_converted_pdf_txt_filename, get_pdf_positions_filename

mutexes = {}


def get_mutex(filename):
    # Ensure that only one thread is accessing a PDF at a time
    if filename not in mutexes:
        mutexes[filename] = Lock()
    return mutexes[filename]


class PDFContent:
    def __init__(self, rawtext, filename, positions):
        self.rawtext = rawtext
        self.filename = filename
        self.positions = positions
        self.pdfium = pdfium.PdfDocument(filename)
        self.mutex = get_mutex(filename)
        self.filetype = "pdf"

    def get_page_image_pil(self, page_number, scale):
        with self.mutex:
            page = self.pdfium[page_number]
            bitmap = page.render(scale=scale)
            return bitmap.to_pil()

    def get_page_chars(self, page_number):
        with self.mutex:
            page = self.pdfium[page_number]
            textpage = page.get_textpage()
            num_chars = textpage.count_chars()
            char_boxes = [textpage.get_charbox(i) for i in range(num_chars)]
            chars = [
                textpage.get_text_range(index=i, count=1) for i in range(num_chars)
            ]
            return [(c, b) for c, b in list(zip(chars, char_boxes))]


# Page separator character
LINE_FEED = "\f"


def get_pdf_content(md5, filename, semantra_dir, force, silent):
    converted_txt = os.path.join(semantra_dir, get_converted_pdf_txt_filename(md5))
    position_index = os.path.join(semantra_dir, get_pdf_positions_filename(md5))

    pdf = pdfium.PdfDocument(filename)
    n_pages = len(pdf)

    if force or not os.path.exists(converted_txt) or not os.path.exists(position_index):
        positions = []
        position = 0
        # newline="" ensures pdfium's \r is preserved
        with open(
            converted_txt, "w", newline="", encoding="utf-8", errors="ignore"
        ) as f:
            for page_index in tqdm(
                range(n_pages),
                desc="Extracting PDF contents",
                leave=False,
                disable=silent,
            ):
                page = pdf[page_index]
                page_width, page_height = page.get_size()
                textpage = page.get_textpage()
                pagetext = textpage.get_text_range()

                positions.append(
                    {
                        "char_index": position,
                        "page_width": page_width,
                        "page_height": page_height,
                    }
                )
                position += f.write(pagetext)
                position += f.write(LINE_FEED)
        with open(position_index, "w", encoding="utf-8") as f:
            json.dump(positions, f)
        with open(
            converted_txt, "r", newline="", encoding="utf-8", errors="ignore"
        ) as f:
            rawtext = f.read()
        return PDFContent(rawtext, filename, positions)
    else:
        with open(
            converted_txt, "r", newline="", encoding="utf-8", errors="ignore"
        ) as f:
            rawtext = f.read()
        with open(position_index, "r", encoding="utf-8") as f:
            positions = json.load(f)

        return PDFContent(rawtext, filename, positions)
